/* SPDX-License-Identifier: GPL-2.0 */
/* Copyright (C) 2004  PaX Team <pageexec@freemail.hu> */

#ifndef _ASM_X86_KERNEXEC_H
#define _ASM_X86_KERNEXEC_H

#ifndef __ASSEMBLY__
#include <linux/compiler.h>
#include <linux/preempt.h>
#include <asm/bug.h>
#include <asm/special_insns.h>

#ifdef CONFIG_PAX_KERNEXEC
#ifdef CONFIG_X86_32
static inline unsigned long __turn_off_size_overflow ktla_ktva(unsigned long addr)
{
	BUG_ON(addr >= PAGE_OFFSET);
	return addr + LOAD_PHYSICAL_ADDR + PAGE_OFFSET;
}
#define ktla_ktva	ktla_ktva

static inline unsigned long __turn_off_size_overflow ktva_ktla(unsigned long addr)
{
	BUG_ON(addr < PAGE_OFFSET);
	return addr - LOAD_PHYSICAL_ADDR - PAGE_OFFSET;
}
#define ktva_ktla	ktva_ktla
#else
#define ktla_ktva(addr)		(addr)
#define ktva_ktla(addr)		(addr)
#endif

static __always_inline unsigned long native_pax_open_kernel(void)
{
	unsigned long cr0;

	preempt_disable();
	barrier();
	// TODO: disable CR4.CET
	cr0 = read_cr0() ^ X86_CR0_WP;
	BUG_ON(cr0 & X86_CR0_WP);
	write_cr0(cr0);
	barrier();
	return cr0 ^ X86_CR0_WP;
}

static __always_inline unsigned long native_pax_close_kernel(void)
{
	unsigned long cr0;

	barrier();
	cr0 = read_cr0() ^ X86_CR0_WP;
	BUG_ON(!(cr0 & X86_CR0_WP));
	write_cr0(cr0);
	// TODO: enable CR4.CET
	barrier();
	preempt_enable_no_resched();
	return cr0 ^ X86_CR0_WP;
}

#ifdef CONFIG_PARAVIRT_XXL
#include <asm/paravirt.h>
#else  /* !CONFIG_PARAVIRT_XXL */
static __always_inline unsigned long pax_open_kernel(void) __opens(kernel) __maybe_used;
static __always_inline unsigned long pax_open_kernel(void)
{
	__open(kernel);
	return native_pax_open_kernel();
}
#define __HAVE_ARCH_PAX_OPEN_KERNEL

static __always_inline unsigned long pax_close_kernel(void) __closes(kernel) __maybe_used;
static __always_inline unsigned long pax_close_kernel(void)
{
	__close(kernel);
	return native_pax_close_kernel();
}
#define __HAVE_ARCH_PAX_CLOSE_KERNEL
#endif /* CONFIG_PARAVIRT_XXL */

#endif /* CONFIG_PAX_KERNEXEC */

#include <asm-generic/kernexec.h>

#endif /* __ASSEMBLY__ */

#endif /* _ASM_X86_KERNEXEC_H */
