#!/bin/sh
# SPDX-License-Identifier: GPL-2.0

ATOMICDIR=$(dirname $0)

. ${ATOMICDIR}/atomic-tbl.sh

#gen_cast(arg, int, atomic)
gen_cast()
{
	local arg="$1"; shift
	local int="$1"; shift
	local atomic="$1"; shift

	[ "${arg%%:*}" = "p" ] || return

	printf "($(gen_param_type "${arg}" "${int}" "${atomic}"))"
}

#gen_args_cast(int, atomic, arg...)
gen_args_cast()
{
	local int="$1"; shift
	local atomic="$1"; shift

	while [ "$#" -gt 0 ]; do
		local cast="$(gen_cast "$1" "${int}" "${atomic}")"
		local arg="$(gen_param_name "$1")"
		printf "${cast}${arg}"
		[ "$#" -gt 1 ] && printf ", "
		shift;
	done
}

#gen_proto_order_variant(meta, pfx, name, sfx, unchecked, order, atomic, int, arg...)
gen_proto_order_variant()
{
	local meta="$1"; shift
	local name="$1$2$3$4$5"; shift; shift; shift; shift; shift
	local atomic="$1"; shift
	local int="$1"; shift

	local ret="$(gen_ret_type "${meta}" "long")"
	local params="$(gen_params "long" "atomic_long${unchecked}" "$@")"
	local argscast="$(gen_args_cast "${int}" "${atomic}${unchecked}" "$@")"
	local retstmt="$(gen_ret_stmt "${meta}")"

cat <<EOF
static inline ${ret} __turn_off_size_overflow
atomic_long_${name}(${params})
{
	${retstmt}${atomic}_${name}(${argscast});
}

EOF
}

cat << EOF
// SPDX-License-Identifier: GPL-2.0
// Copyright (C) 2009  PaX Team <pageexec@freemail.hu>

// Generated by $0
// DO NOT MODIFY THIS FILE DIRECTLY

#ifndef _ASM_GENERIC_ATOMIC_LONG_H
#define _ASM_GENERIC_ATOMIC_LONG_H

#include <asm/types.h>

#ifdef CONFIG_64BIT
typedef atomic64_t atomic_long_t;

#ifdef CONFIG_PAX_REFCOUNT
typedef atomic64_unchecked_t atomic_long_unchecked_t;
#else
typedef atomic64_t atomic_long_unchecked_t;
#endif

#define ATOMIC_LONG_INIT(i)		ATOMIC64_INIT(i)
#define atomic_long_cond_read_acquire	atomic64_cond_read_acquire
#define atomic_long_cond_read_relaxed	atomic64_cond_read_relaxed
#else
typedef atomic_t atomic_long_t;

#ifdef CONFIG_PAX_REFCOUNT
typedef atomic_unchecked_t atomic_long_unchecked_t;
#else
typedef atomic_t atomic_long_unchecked_t;
#endif

#define ATOMIC_LONG_INIT(i)		ATOMIC_INIT(i)
#define atomic_long_cond_read_acquire	atomic_cond_read_acquire
#define atomic_long_cond_read_relaxed	atomic_cond_read_relaxed
#endif

#ifdef CONFIG_64BIT

EOF

grep '^[aU]' "$1" | while read kind name meta args; do
	gen_proto "${kind}" "${meta}" "${name}" "atomic64" "s64" ${args}
done

cat <<EOF
#else /* CONFIG_64BIT */

EOF

grep '^[au]' "$1" | while read kind name meta args; do
	gen_proto "${kind}" "${meta}" "${name}" "atomic" "int" ${args}
done

cat <<EOF
#endif /* CONFIG_64BIT */

#ifdef CONFIG_PAX_REFCOUNT
static inline void pax_refcount_needs_these_functions(void)
{
	atomic_read_unchecked((atomic_unchecked_t *)NULL);
	atomic_set_unchecked((atomic_unchecked_t *)NULL, 0);
	atomic_add_unchecked(0, (atomic_unchecked_t *)NULL);
	atomic_sub_unchecked(0, (atomic_unchecked_t *)NULL);
	atomic_inc_unchecked((atomic_unchecked_t *)NULL);
	(void)atomic_inc_and_test_unchecked((atomic_unchecked_t *)NULL);
	atomic_inc_return_unchecked((atomic_unchecked_t *)NULL);
	atomic_fetch_inc_unchecked((atomic_unchecked_t *)NULL);
	atomic_add_return_unchecked(0, (atomic_unchecked_t *)NULL);
	atomic_sub_return_unchecked(0, (atomic_unchecked_t *)NULL);
	atomic_dec_unchecked((atomic_unchecked_t *)NULL);
	atomic_dec_return_unchecked((atomic_unchecked_t *)NULL);
	atomic_cmpxchg_unchecked((atomic_unchecked_t *)NULL, 0, 0);
	(void)atomic_xchg_unchecked((atomic_unchecked_t *)NULL, 0);
	atomic_andnot_unchecked(0, (atomic_unchecked_t *)NULL);

	atomic_long_read_unchecked((atomic_long_unchecked_t *)NULL);
	atomic_long_set_unchecked((atomic_long_unchecked_t *)NULL, 0);
	atomic_long_add_unchecked(0, (atomic_long_unchecked_t *)NULL);
	atomic_long_sub_unchecked(0, (atomic_long_unchecked_t *)NULL);
	atomic_long_inc_unchecked((atomic_long_unchecked_t *)NULL);
	atomic_long_add_return_unchecked(0, (atomic_long_unchecked_t *)NULL);
	atomic_long_sub_return_unchecked(0, (atomic_long_unchecked_t *)NULL);
	atomic_long_inc_return_unchecked((atomic_long_unchecked_t *)NULL);
	atomic_long_fetch_inc_unchecked((atomic_long_unchecked_t *)NULL);
	atomic_long_dec_unchecked((atomic_long_unchecked_t *)NULL);
	atomic_long_dec_return_unchecked((atomic_long_unchecked_t *)NULL);
	(void)atomic_long_cmpxchg_unchecked((atomic_long_unchecked_t *)NULL, 0, 0);
}
#else /* !CONFIG_PAX_REFCOUNT */

#ifndef atomic_read_unchecked
#define atomic_read_unchecked(v) atomic_read(v)
#endif

#ifndef atomic_set_unchecked
#define atomic_set_unchecked(v, i) atomic_set((v), (i))
#endif

#ifndef atomic_add_unchecked
#define atomic_add_unchecked(i, v) atomic_add((i), (v))
#endif

#ifndef atomic_sub_unchecked
#define atomic_sub_unchecked(i, v) atomic_sub((i), (v))
#endif

#ifndef atomic_inc_unchecked
#define atomic_inc_unchecked(v) atomic_inc(v)
#endif

#ifndef atomic_inc_and_test_unchecked
#define atomic_inc_and_test_unchecked(v) atomic_inc_and_test(v)
#endif

#ifndef atomic_inc_return_unchecked
#define atomic_inc_return_unchecked(v) atomic_inc_return(v)
#endif

#ifndef atomic_fetch_inc_unchecked
#define atomic_fetch_inc_unchecked(v) atomic_fetch_inc(v)
#endif

#ifndef atomic_add_return_unchecked
#define atomic_add_return_unchecked(i, v) atomic_add_return((i), (v))
#endif

#ifndef atomic_sub_return_unchecked
#define atomic_sub_return_unchecked(i, v) atomic_sub_return((i), (v))
#endif

#ifndef atomic_dec_unchecked
#define atomic_dec_unchecked(v) atomic_dec(v)
#endif

#ifndef atomic_dec_return_unchecked
#define atomic_dec_return_unchecked(v) atomic_dec_return(v)
#endif

#ifndef atomic_cmpxchg_unchecked
#define atomic_cmpxchg_unchecked(v, o, n) atomic_cmpxchg((v), (o), (n))
#endif

#ifndef atomic_xchg_unchecked
#define atomic_xchg_unchecked(v, i) atomic_xchg((v), (i))
#endif

#ifndef atomic_andnot_unchecked
#define atomic_andnot_unchecked(i, v) atomic_andnot((i), (v))
#endif

#define atomic_long_read_unchecked(v) atomic_long_read(v)
#define atomic_long_set_unchecked(v, i) atomic_long_set((v), (i))
#define atomic_long_add_unchecked(i, v) atomic_long_add((i), (v))
#define atomic_long_sub_unchecked(i, v) atomic_long_sub((i), (v))
#define atomic_long_inc_unchecked(v) atomic_long_inc(v)
#define atomic_long_fetch_inc_unchecked(v) atomic_long_fetch_inc(v)
#define atomic_long_add_return_unchecked(i, v) atomic_long_add_return((i), (v))
#define atomic_long_sub_return_unchecked(i, v) atomic_long_sub_return((i), (v))
#define atomic_long_inc_return_unchecked(v) atomic_long_inc_return(v)
#define atomic_long_dec_unchecked(v) atomic_long_dec(v)
#define atomic_long_dec_return_unchecked(v) atomic_long_dec_return(v)

#ifndef atomic_long_xchg_unchecked
#define atomic_long_xchg_unchecked(v, i) atomic_long_xchg((v), (i))
#endif

#ifndef atomic_long_cmpxchg_unchecked
#define atomic_long_cmpxchg_unchecked(v, o, n) atomic_long_cmpxchg((v), (o), (n))
#endif

#endif /* CONFIG_PAX_REFCOUNT */

#endif /* _ASM_GENERIC_ATOMIC_LONG_H */
EOF
