/*
 * Copyright 2019-2020 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * dump the current function into the gimple dump file as gcc 8 and newer do it
 */

#include "utilities.h"

#if BUILDING_GCC_VERSION < 8000
static void execute_function_dump(struct function *fn, void *data ATTRIBUTE_UNUSED)
{
	if (!dump_file)
		return;

	push_cfun(fn);

	if (fn->curr_properties & PROP_trees)
		dump_function_to_file(fn->decl, dump_file, dump_flags);
	else
		print_rtl_with_bb(dump_file, get_insns(), dump_flags);

	fflush(dump_file);

	if ((cfun->curr_properties & PROP_cfg) && (dump_flags & TDF_GRAPH))
#if BUILDING_GCC_VERSION < 4008
		if (!(fn->curr_properties & PROP_trees))
			print_rtl_graph_with_bb(dump_file_name, get_insns());
#else
		print_graph_cfg(dump_file_name, cfun);
#endif

	pop_cfun();
}

#if BUILDING_GCC_VERSION < 4006
void (*old_internal_error)(const char *, va_list *);

void emergency_dump_internal_error_function(const char *msgid, va_list *ap)
#else
void (*old_internal_error)(diagnostic_context *, const char *, va_list *);

void emergency_dump_internal_error_function(diagnostic_context *dc, const char *msgid, va_list *ap)
#endif
{
	enum opt_pass_type pt;

	if (old_internal_error)
#if BUILDING_GCC_VERSION < 4006
		old_internal_error(msgid, ap);
#else
		old_internal_error(dc, msgid, ap);
#endif

	if (!current_pass || !dump_file || !cfun)
		return;

	pt = current_pass->type;
	fnotice(stderr, "during %s pass: %s\n", pt == GIMPLE_PASS ? "GIMPLE" : pt == RTL_PASS ? "RTL" : "IPA", current_pass->name);
	fnotice(stderr, "dump file: %s\n", dump_file_name);
	fprintf(dump_file, "\n\n\nEMERGENCY DUMP:\n\n");
	execute_function_dump(cfun, current_pass);
}
#endif
